(*  Title:      Nitpick/Nitpick.thy
    Author:     Jasmin Blanchette, TU Muenchen
    Copyright   2008, 2009

Nitpick: Yet another counterexample generator for Isabelle/HOL.
*)

header {* Nitpick *}

theory Nitpick
imports Main
uses ("Tools/kodkod.ML")
     ("Tools/kodkod_sat.ML")
     ("Tools/nitpick_util.ML")
     ("Tools/nitpick_hol.ML")
     ("Tools/nitpick_scope.ML")
     ("Tools/nitpick_peephole.ML")
     ("Tools/nitpick_rep.ML")
     ("Tools/nitpick_term.ML")
     ("Tools/nitpick_kodkod.ML")
     ("Tools/nitpick_model.ML")
     ("Tools/nitpick.ML")
     ("Tools/nitpick_isar.ML")
     ("Tools/nitpick_tests.ML")
begin

text {*
The following lemmas are not strictly necessary but they help the
"specialize_funs" optimization.
*}

lemma The_psimp [nitpick_const_psimp]:
"P = {x} \<Longrightarrow> The P = x"
by (metis Collect_def singleton_conv2 the_sym_eq_trivial)

lemma Eps_psimp [nitpick_const_psimp]:
"\<lbrakk>P x; \<not> P y; Eps P = y\<rbrakk> \<Longrightarrow> Eps P = x"
by (metis exE_some)

lemma option_case_def [nitpick_const_def]:
"option_case x f p \<equiv> if p = None then x else f (the p)"
apply (subgoal_tac "option_case x f p = (if p = None then x else f (the p))")
 defer
by (case_tac p) auto

lemma nat_case_def [nitpick_const_def]:
"nat_case x f n \<equiv> if n = 0 then x else f (n - 1)"
apply (subgoal_tac "nat_case x f n = (if n = 0 then x else f (n - 1))")
 defer
by (case_tac n) auto

lemma list_case_def [nitpick_const_def]:
"list_case x f ys \<equiv> if ys = [] then x else f (hd ys) (tl ys)"
apply (subgoal_tac "list_case x f ys =
                    (if ys = [] then x else f (hd ys) (tl ys))")
 defer
by (case_tac ys) auto

declare list.size(3-4) [nitpick_const_simp del]

lemma list_size_simp [nitpick_const_simp]:
"size xs = (if xs = [] then 0 else Suc (size (tl xs)))"
by (case_tac xs) auto

use "Tools/kodkod.ML"
use "Tools/kodkod_sat.ML"
use "Tools/nitpick_util.ML"
use "Tools/nitpick_hol.ML"
use "Tools/nitpick_scope.ML"
use "Tools/nitpick_peephole.ML"
use "Tools/nitpick_rep.ML"
use "Tools/nitpick_term.ML"
use "Tools/nitpick_kodkod.ML"
use "Tools/nitpick_model.ML"
use "Tools/nitpick.ML"
use "Tools/nitpick_isar.ML"
use "Tools/nitpick_tests.ML"

end
